import warnings
from typing import Callable, Optional

import numpy as np

from CADETProcess import CADETProcessError, SimulationResults
from CADETProcess.fractionation import Fractionator
from CADETProcess.optimization import (
    COBYLA,
    OptimizationProblem,
    OptimizationResults,
    OptimizerBase,
)
from CADETProcess.performance import Mass, Performance, Purity

__all__ = ["FractionationOptimizer"]


class FractionationEvaluator:
    """Dummy Evaluator to enable caching."""

    def evaluate(self, fractionator: Fractionator) -> Performance:
        """
        Evaluate the fractionator.

        Parameters
        ----------
        fractionator: Fractionator
            The Fractionator object to be evaluated.

        Returns
        -------
        object
            The evaluation result.
        """
        return fractionator.performance

    __call__ = evaluate

    def __str__(self) -> str:
        """str: Name of the FractionationEvaluator."""
        return __class__.__name__


class FractionationOptimizer:
    """Configuration for fractionating Chromatograms."""

    def __init__(
        self,
        optimizer: Optional[OptimizerBase] = None,
        log_level: str = "WARNING",
    ) -> None:
        """
        Initialize the FractionationOptimizer.

        Parameters
        ----------
        optimizer: OptimizerBase, optional
            Optimizer for optimizing the fractionation times.
            If no value is specified, a default COBYLA optimizer will be used.
        log_level: {'WARNING', 'INFO', 'DEBUG', 'ERROR'}
            Log level for the fractionation optimization process.
            The default is 'WARNING'.
        """
        if optimizer is None:
            optimizer = COBYLA()
            optimizer.tol = 1e-4
            optimizer.catol = 5e-3
            optimizer.rhobeg = 1e-3

        self.optimizer = optimizer
        self.log_level = log_level

    @property
    def optimizer(self) -> OptimizerBase:
        """OptimizerBase: Optimizer for optimizing the fractionation times."""
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer: OptimizerBase) -> None:
        """
        Set the optimizer.

        Parameters
        ----------
        optimizer: OptimizerBase
            The optimizer to be set.

        Raises
        ------
        TypeError
            If the optimizer is not an instance of OptimizerBase.
        """
        if not isinstance(optimizer, OptimizerBase):
            raise TypeError("Expected OptimizerBase")
        self._optimizer = optimizer

    def _setup_fractionator(
        self,
        simulation_results: SimulationResults,
        purity_required: list[float],
        components: Optional[list] = None,
        use_total_concentration_components: bool = True,
        allow_empty_fractions: bool = True,
    ) -> Fractionator:
        """
        Set up the Fractionator for optimizing the fractionation times of Chromatograms.

        Parameters
        ----------
        simulation_results: object
            Simulation results to be used for setting up the Fractionator object.
        purity_required : list of floats
            Minimum purity required for the components in the fractionation.
        components: list, optional
            List of components to consider in the fractionation process.
        use_total_concentration_components: bool, optional
            If True, use the total concentration of the components. The default is True.
        allow_empty_fractions: bool, optional
            If True, allow empty fractions. The default is True.

        Returns
        -------
        Fractionator
            The Fractionator object that has been set up using the provided arguments.
        """
        frac = Fractionator(
            simulation_results,
            components=components,
            use_total_concentration_components=use_total_concentration_components,
        )

        frac.initial_values(purity_required)

        if not np.any(frac.n_fractions_per_pool[:-1]):
            raise CADETProcessError("No areas found with sufficient purity.")

        if not allow_empty_fractions:
            empty_fractions = []
            for i, comp in enumerate(purity_required):
                if comp > 0:
                    if frac.fraction_pools[i].n_fractions == 0:
                        empty_fractions.append(i)
            if len(empty_fractions) != 0:
                raise CADETProcessError(
                    "No areas found with sufficient purity for component(s) "
                    f"{[str(frac.component_system[i]) for i in empty_fractions]}."
                )

        return frac

    def _setup_optimization_problem(
        self,
        frac: Fractionator,
        purity_required: list[float],
        allow_empty_fractions: bool = True,
        ranking: Optional[int | list[float]] = 1,
        obj_fun: Optional[Callable] = None,
        minimize: bool = True,
        bad_metrics: Optional[float | list[float]] = None,
        n_objectives: int = 1,
    ) -> tuple[OptimizationProblem, list[float]]:
        """
        Set up the OptimizationProblem for optimizing the fractionation times.

        Parameters
        ----------
        frac : Fractionator
            The Fractionator object.
        purity_required : list[float]
            Minimum purity required for the components in the fractionation.
        allow_empty_fractions: bool, optional
            If True, allow empty fractions. The default is True.
        ranking : Optional[int | list[float]] = 1,
            Weighting factors for individual components.
            If 1, the same value is assumed for all components.
            If None, no ranking is used and the problem is solved as multi-objective.
            The default is 1.
        obj_fun : callable, optional
            Alternative objective function.
            If no function is provided, the fraction mass is maximized.
            The default is None.
        bad_metrics : float or list of floats, optional
            Values to be returned if evaluation of objective function failes.
            The default is 0.
        minimize : bool, optional
            If True, the obj_fun is assumed to return a value that is to be minimized.
            The default it True.
        n_objectives : int
            Number of objectives. The default is 1.

        Raises
        ------
        CADETProcessError
            If the optimization problem setup fails.

        Returns
        -------
        OptimizationProblem
            The configured OptimizationProblem object.
        list
            The initial values for the optimization variables.
        """
        # Handle empty fractions
        n_fractions = np.array([pool.n_fractions for pool in frac.fraction_pools])
        empty_fractions = np.where(n_fractions[0:-1] == 0)[0]
        if len(empty_fractions) > 0 and allow_empty_fractions:
            for empty_fraction in empty_fractions:
                purity_required[empty_fraction] = 0

        # Setup Optimization Problem
        opt = OptimizationProblem(
            "FractionationOptimization",
            log_level=self.log_level,
            use_diskcache=False,
        )

        opt.add_evaluation_object(frac)

        frac_evaluator = FractionationEvaluator()
        opt.add_evaluator(frac_evaluator)

        if obj_fun is None:
            obj_fun = Mass(ranking=ranking)
            minimize = False
            bad_metrics = 0

        opt.add_objective(
            obj_fun,
            requires=frac_evaluator,
            n_objectives=n_objectives,
            minimize=minimize,
            bad_metrics=bad_metrics,
        )

        purity = Purity()
        purity.n_metrics = frac.component_system.n_comp
        opt.add_nonlinear_constraint(
            purity,
            n_nonlinear_constraints=len(purity_required),
            bounds=purity_required,
            comparison_operator="ge",
            requires=frac_evaluator,
        )

        for evt in frac.events:
            opt.add_variable(
                evt.name,
                parameter_path=evt.name + ".time",
                lb=-frac.cycle_time,
                ub=2 * frac.cycle_time,
                transform="linear",
            )

        for chrom_index, chrom in enumerate(frac.chromatograms):
            chrom_events = frac.chromatogram_events[chrom]
            evt_names = [evt.name for evt in chrom_events]
            for evt_index, evt in enumerate(chrom_events):
                if evt_index < len(chrom_events) - 1:
                    opt.add_linear_constraint(
                        [evt_names[evt_index], evt_names[evt_index + 1]], [1, -1]
                    )
                else:
                    opt.add_linear_constraint(
                        [evt_names[0], evt_names[-1]], [-1, 1], frac.cycle_time
                    )

        x0 = [evt.time for evt in frac.events]

        if not opt.check_nonlinear_constraints(x0):
            raise CADETProcessError("No areas found with sufficient purity.")

        return opt, x0

    def optimize_fractionation(
        self,
        simulation_results: SimulationResults,
        purity_required: float | list[float],
        components: Optional[list[str]] = None,
        use_total_concentration_components: bool = True,
        ranking: Optional[int | list[float]] = 1,
        obj_fun: Optional[Callable] = None,
        n_objectives: int = 1,
        bad_metrics: float | list[float] = 0,
        minimize: bool = True,
        allow_empty_fractions: bool = True,
        ignore_failed: bool = False,
        return_optimization_results: bool = False,
        save_results: bool = False,
    ) -> Fractionator | tuple[Fractionator, OptimizationResults]:
        """
        Optimize the fractionation times with respect to purity constraints.

        Parameters
        ----------
        simulation_results : SimulationResults
            Results containing the chromatograms for fractionation.
        purity_required :  float or array_like
            Minimum required purity for components. If is float, the same
            value is assumed for all components.
        components : list
            List of components to consider in the fractionation process.
        use_total_concentration_components : bool, Default=True
            Flag wheter to use the total concentration components.
        ranking : Optional[int | list[float]] = 1,
            Weighting factors for individual components.
            If 1, the same value is assumed for all components.
            If None, no ranking is used and the problem is solved as multi-objective.
            The default is 1.
        obj_fun : function, optional
            Objective function used for OptimizationProblem.
            If COBYLA is used, must return single objective.
            If is None, the mass of all components is maximized.
        n_objectives : int, optional
            Number of objectives returned by obj_fun. The default is 1.
        bad_metrics : float or list of floats, optional
            Values to be returned if evaluation of objective function failes.
            The default is 0.
        minimize : bool, optional
            If True, the obj_fun is assumed to return a value that is to be minimized.
            The default it True.
        allow_empty_fractions: bool, optional
            If True, allow empty fractions. The default is True.
        ignore_failed : bool, optional
            Ignore failed optimization and use initial values.
            The default is False.
        return_optimization_results : bool, optional
            If True, return optimization results.
            Otherwise, return fractionation object.
            The default is False.
        save_results : bool, optional
            If True, save optimization results. The default is False.

        Returns
        -------
        Fractionator or OptimizationResults
            The Fractionator object with optimized cut times
            or the OptimizationResults object.

        Raises
        ------
        TypeError
            If simulation_results is not an instance of SimulationResults.
        CADETProcessError
            If simulation_results do not contain chromatograms.
        Warning
            If purity requirements cannot be fulfilled.

        See Also
        --------
        _setup_fractionator
        _setup_optimization_problem
        Fractionator
        CADETProcess.solution.SolutionIO
        CADETProcess.optimization.OptimizationProblem
        CADETProcess.optimization.OptimizerBase
        """
        if not isinstance(simulation_results, SimulationResults):
            raise TypeError("Expected SimulationResults.")

        if len(simulation_results.chromatograms) == 0:
            raise CADETProcessError("Simulation results do not contain chromatogram.")

        if isinstance(purity_required, float):
            n_comp = simulation_results.component_system.n_comp
            purity_required = n_comp * [purity_required]

        # Store previous lock state, unlock to ensure consistent values
        lock_state = simulation_results.process.lock
        simulation_results.process.lock = False

        frac = self._setup_fractionator(
            simulation_results,
            purity_required,
            components=components,
            use_total_concentration_components=use_total_concentration_components,
            allow_empty_fractions=allow_empty_fractions,
        )

        opt, x0 = self._setup_optimization_problem(
            frac,
            purity_required,
            allow_empty_fractions=allow_empty_fractions,
            ranking=ranking,
            obj_fun=obj_fun,
            n_objectives=n_objectives,
            minimize=minimize,
            bad_metrics=bad_metrics,
        )

        # Lock to enable caching
        simulation_results.process.lock = True
        try:
            results = self.optimizer.optimize(
                opt,
                x0,
                save_results=save_results,
                log_level=self.log_level,
                delete_cache=True,
            )
            opt.set_variables(results.x[0])
            frac.reset()
        except CADETProcessError as e:
            if ignore_failed:
                warnings.warn("Optimization failed. Returning initial values")
                frac.initial_values(purity_required)
            else:
                raise CADETProcessError(str(e))
        finally:
            # Restore previous lock state
            simulation_results.process.lock = lock_state

        if return_optimization_results:
            return results
        else:
            return frac

    evaluate = optimize_fractionation
    __call__ = evaluate

    def __str__(self) -> str:
        """Name of the FractionationOptimizer."""
        return self.__class__.__name__



"""

To do
=====

## SolutionBase
- [x] Total concentration
- [x] Total concentration components
- [x] Local Purity
- [x] Local Purity Components

## SolutionIO
- [x] Mass Flow
- [x] Fraction Mass
- [x] Fraction Volume
- [ ] Interpolation
- [ ] Resampling
- [ ] Normalization
- [ ] Smoothing
- [ ] Derivative
- [ ] plotting?

## SolutionBulk
- [ ] Total concentration
- [ ] Plotting

Notes
-----
To facilitate trouble shooting, make use of built-in plot functions.
They are available for the Solution, as well as the TimeLine objects.

"""

import unittest

import numpy as np
from CADETProcess.dynamicEvents import Section, TimeLine
from CADETProcess.processModel import ComponentSystem
from CADETProcess.solution import SolutionIO, slice_solution
from scipy import stats

show_plots = False

comp_2 = ComponentSystem(2)

comp_2_3_species = ComponentSystem()
comp_2_3_species.add_component("A")
comp_2_3_species.add_component("B", species=["B+", "B-"])

# Note that signal is recorded at 10 Hz.
time = np.linspace(0, 100, 1001)

solution_2_constant = np.ones((len(time), 2))
solution_2_constant[:, 1] *= 2

solution_2_square = np.zeros((len(time), 2))
solution_2_square[200:400, 0] = 2
solution_2_square[300:500, 1] = 1

solution_2_gaussian = np.zeros((len(time), 2))
mu_0 = 30
sigma_0 = 5
solution_2_gaussian[:, 0] = stats.norm.pdf(time, mu_0, sigma_0)
mu_1 = 40
sigma_1 = 5
solution_2_gaussian[:, 1] = stats.norm.pdf(time, mu_1, sigma_1)

solution_3_linear = np.zeros((len(time), 3))

solution_3_linear[100:201, 0] = 1
solution_3_linear[200:401, 0] = np.linspace(1, 0, 201)

solution_3_linear[200:401, 1] = np.linspace(0, 1, 201)
solution_3_linear[400:601, 1] = np.linspace(1, 0, 201)

solution_3_linear[300:501, 2] = np.linspace(0, 1, 201)
solution_3_linear[500:701, 2] = np.linspace(1, 0, 201)

q_const = np.ones(time.shape)

q_interrupted = TimeLine()
q_interrupted.add_section(Section(0, 30, [1, 0, 0, 0], is_polynomial=True))
q_interrupted.add_section(Section(30, 40, [0, 0, 0, 0], is_polynomial=True))
q_interrupted.add_section(Section(40, 100, [1, 0, 0, 0], is_polynomial=True))

q_linear = TimeLine()
q_linear.add_section(Section(0, 35, [0, 0, 0, 0], is_polynomial=True))
q_linear.add_section(Section(35, 45, [0, 1 / 10, 0, 0], is_polynomial=True))
q_linear.add_section(Section(45, 100, [0, 0, 0, 0], is_polynomial=True))


class TestSolution(unittest.TestCase):
    def setUp(self):
        # 2 Components, constant concentration, constant flow
        self.solution_constant = SolutionIO(
            "simple", comp_2, time, solution_2_constant, q_const
        )

        # 2 Components, square peaks, constant flow
        self.solution_square = SolutionIO(
            "simple", comp_2, time, solution_2_square, q_const
        )

        # 3 Components (species), linear peaks, constant flow
        self.solution_species = SolutionIO(
            "simple", comp_2_3_species, time, solution_3_linear, q_const
        )

        # 2 Components, constant concentration, interrupted flow
        self.solution_constant_interrupted = SolutionIO(
            "simple", comp_2, time, solution_2_constant, q_interrupted
        )

        # 2 Components, constant concentration, linear flow
        self.solution_constant_linear = SolutionIO(
            "simple", comp_2, time, solution_2_constant, q_linear
        )

        # 2 Components, square peaks, interrupted flow
        self.solution_square_interrupted = SolutionIO(
            "simple", comp_2, time, solution_2_square, q_interrupted
        )

        # 2 Components, square peaks, linear flow
        self.solution_square_linear = SolutionIO(
            "simple", comp_2, time, solution_2_square, q_linear
        )

        # 2 Components, gaussian peaks, constant flow
        self.solution_gaussian = SolutionIO(
            "simple", comp_2, time, solution_2_gaussian, q_const
        )

        # 2 Components, gaussian peaks, interrupted flow
        self.solution_gaussian_interrupted = SolutionIO(
            "simple", comp_2, time, solution_2_gaussian, q_interrupted
        )

        # 2 Components, gaussian peaks, linear flow
        self.solution_gaussian_linear = SolutionIO(
            "simple", comp_2, time, solution_2_gaussian, q_linear
        )

    def test_total_concentration(self):
        # Simple case
        c_total_expected = 3
        c_total = self.solution_constant.total_concentration
        np.testing.assert_almost_equal(c_total, c_total_expected)

        # 2 Components, square peaks
        c_total = self.solution_square.total_concentration

        # Note that here t is not the acutal time but rather the index.
        t = 0
        c_total_expected = [0]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 250
        c_total_expected = [2]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 350
        c_total_expected = [3]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 450
        c_total_expected = [1]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 550
        c_total_expected = [0]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        # 2 Components, species
        c_total = self.solution_species.total_concentration

        t = 150
        c_total_expected = [1]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 300
        c_total_expected = [1]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 350
        c_total_expected = [1.25]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 450
        c_total_expected = [1.5]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 550
        c_total_expected = [1]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

        t = 650
        c_total_expected = [0.25]
        c_total_actual = c_total[t]
        np.testing.assert_almost_equal(c_total_actual, c_total_expected)

    def test_total_concentration_components(self):
        # 2 Components, species
        c_total_comp = self.solution_species.total_concentration_components

        t = 150
        c_total_comp_expected = [1, 0]
        c_total_comp_actual = c_total_comp[t]
        np.testing.assert_almost_equal(c_total_comp_actual, c_total_comp_expected)

        t = 300
        c_total_comp_expected = [0.5, 0.5]
        c_total_comp_actual = c_total_comp[t]
        np.testing.assert_almost_equal(c_total_comp_actual, c_total_comp_expected)

        t = 350
        c_total_comp_expected = [0.25, 1]
        c_total_comp_actual = c_total_comp[t]
        np.testing.assert_almost_equal(c_total_comp_actual, c_total_comp_expected)

        t = 450
        c_total_comp_expected = [0, 1.5]
        c_total_comp_actual = c_total_comp[t]
        np.testing.assert_almost_equal(c_total_comp_actual, c_total_comp_expected)

    def test_local_purity_components(self):
        # Simple case
        local_purity_expected = [1 / 3, 2 / 3]
        local_purity_actual = self.solution_constant.local_purity_components[0]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        # 2 Components, square peaks
        local_purity = self.solution_square.local_purity_components

        t = 0
        local_purity_expected = [0, 0]
        local_purity_actual = local_purity[t]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        t = 250
        local_purity_expected = [1, 0]
        local_purity_actual = local_purity[t]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        t = 350
        local_purity_expected = [2 / 3, 1 / 3]
        local_purity_actual = local_purity[t]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        t = 450
        local_purity_expected = [0, 1]
        local_purity_actual = local_purity[t]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        t = 550
        local_purity_expected = [0, 0]
        local_purity_actual = local_purity[t]
        np.testing.assert_almost_equal(local_purity_actual, local_purity_expected)

        # 2 Components, 3 species (summed up)
        local_purity_components = self.solution_species.local_purity_components

        t = 150
        local_purity_components_expected = [1, 0]
        local_purity_components_actual = local_purity_components[t]
        np.testing.assert_almost_equal(
            local_purity_components_actual, local_purity_components_expected
        )

        t = 300
        local_purity_components_expected = [0.5, 0.5]
        local_purity_components_actual = local_purity_components[t]
        np.testing.assert_almost_equal(
            local_purity_components_actual, local_purity_components_expected
        )

        t = 350
        local_purity_components_expected = [0.2, 0.8]
        local_purity_components_actual = local_purity_components[t]
        np.testing.assert_almost_equal(
            local_purity_components_actual, local_purity_components_expected
        )

        t = 450
        local_purity_components_expected = [0, 1]
        local_purity_components_actual = local_purity_components[t]
        np.testing.assert_almost_equal(
            local_purity_components_actual, local_purity_components_expected
        )

    def test_local_purity_species(self):
        # 2 Components, 3  species
        local_purity_species = self.solution_species.local_purity_species

        t = 150
        local_purity_species_expected = [1, 0, 0]
        local_purity_species_actual = local_purity_species[t]
        np.testing.assert_almost_equal(
            local_purity_species_actual, local_purity_species_expected
        )

        t = 300
        local_purity_species_expected = [0.5, 0.5, 0]
        local_purity_species_actual = local_purity_species[t]
        np.testing.assert_almost_equal(
            local_purity_species_actual, local_purity_species_expected
        )

        t = 350
        local_purity_species_expected = [0.2, 0.6, 0.2]
        local_purity_species_actual = local_purity_species[t]
        np.testing.assert_almost_equal(
            local_purity_species_actual, local_purity_species_expected
        )

        t = 450
        local_purity_species_expected = [0, 0.5, 0.5]
        local_purity_species_actual = local_purity_species[t]
        np.testing.assert_almost_equal(
            local_purity_species_actual, local_purity_species_expected
        )

    def test_resampling(self):
        pass

    def test_normalization(self):
        pass

    def test_smoothing(self):
        pass

    def test_fraction_volume(self):
        volume_expected = 0
        volume_actual = self.solution_constant.fraction_volume(0, 0)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 10
        volume_actual = self.solution_constant.fraction_volume(0, 10)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 0
        volume_actual = self.solution_constant_interrupted.fraction_volume(30, 40)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 20
        volume_actual = self.solution_constant_interrupted.fraction_volume(20, 50)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 0
        volume_actual = self.solution_constant_linear.fraction_volume(0, 10)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 5
        volume_actual = self.solution_constant_linear.fraction_volume(25, 45)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

        volume_expected = 1.25
        volume_actual = self.solution_constant_linear.fraction_volume(30, 40)
        np.testing.assert_almost_equal(volume_actual, volume_expected)

    def test_fraction_mass(self):
        mass_expected = [10, 20]
        mass_actual = self.solution_constant.fraction_mass(0, 10)
        np.testing.assert_almost_equal(mass_actual, mass_expected)

        mass_expected = [10, 20]
        mass_actual = self.solution_constant.fraction_mass(0, 10)
        np.testing.assert_almost_equal(mass_actual, mass_expected)

        mass_expected = np.array(
            [np.diff(stats.norm.cdf([-2, 2]))[0], np.diff(stats.norm.cdf([-4, 0]))[0]]
        )
        mass_actual = self.solution_gaussian.fraction_mass(20, 40)
        np.testing.assert_almost_equal(mass_actual, mass_expected)

        mass_expected = np.array(
            [np.diff(stats.norm.cdf([-2, 0]))[0], np.diff(stats.norm.cdf([-4, -2]))[0]]
        )
        mass_actual = self.solution_gaussian_interrupted.fraction_mass(20, 40)
        np.testing.assert_almost_equal(mass_actual, mass_expected)


class TestSliceSolution(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        self.solution_gaussian = SolutionIO(
            "simple", comp_2, time, solution_2_gaussian, q_const
        )

        self.solution_species = SolutionIO(
            "simple", comp_2_3_species, time, solution_3_linear, q_const
        )

    def test_coordinates(self):
        solution = slice_solution(
            self.solution_gaussian, coordinates={"time": [30, 40]}
        )

        np.testing.assert_almost_equal(solution.time, np.linspace(30, 40, 101))
        np.testing.assert_almost_equal(
            self.solution_gaussian.solution[300:401, ...], solution.solution
        )

    def test_components(self):
        # Test single component
        solution = slice_solution(self.solution_gaussian, components="0")
        self.assertEqual(solution.component_system.names, ["0"])

        np.testing.assert_equal(
            self.solution_gaussian.solution[..., [0]], solution.solution
        )

        # Test with multiple components

        # Test with component that has > 1 species
        solution = slice_solution(self.solution_species, components="B")
        self.assertEqual(solution.component_system.names, ["B"])
        self.assertEqual(solution.component_system.species, ["B+", "B-"])

        np.testing.assert_equal(
            self.solution_species.solution[..., 1:], solution.solution
        )

    def test_component_concentration(self):
        solution = slice_solution(
            self.solution_species, use_total_concentration_components=True
        )
        self.assertEqual(solution.component_system.names, ["A", "B"])
        self.assertEqual(solution.component_system.species, ["A", "B"])

        np.testing.assert_equal(
            self.solution_species.solution[..., 0], solution.solution[:, 0]
        )

        np.testing.assert_equal(
            np.sum(self.solution_species.solution[..., 1:]),
            np.sum(solution.solution[:, 1]),
        )
        np.testing.assert_equal(
            self.solution_species.total_concentration_components,
            solution.total_concentration_components,
        )

    def test_total_concentration(self):
        solution = slice_solution(self.solution_species, use_total_concentration=True)
        self.assertEqual(solution.component_system.names, ["total_concentration"])

        np.testing.assert_equal(
            np.sum(self.solution_species.solution[..., :]),
            np.sum(solution.solution, axis=0),
        )
        np.testing.assert_equal(
            self.solution_species.total_concentration, solution.total_concentration
        )

        # One component
        solution = slice_solution(
            self.solution_species, components=["A"], use_total_concentration=True
        )
        self.assertEqual(solution.component_system.names, ["total_concentration"])

        np.testing.assert_equal(
            self.solution_species.solution[..., [0]], solution.solution
        )

        np.testing.assert_equal(
            self.solution_species.solution[..., [0]], solution.total_concentration
        )

        # One component with two species
        solution = slice_solution(
            self.solution_species, components=["B"], use_total_concentration=True
        )
        self.assertEqual(solution.component_system.names, ["total_concentration"])

        np.testing.assert_equal(
            np.sum(self.solution_species.solution[..., 1:], keepdims=True, axis=1),
            solution.solution,
        )

        np.testing.assert_equal(
            np.sum(self.solution_species.solution[..., 1:], keepdims=True, axis=1),
            solution.total_concentration,
        )


class TestPlot(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        self.solution_species = SolutionIO(
            "simple", comp_2_3_species, time, solution_3_linear, q_const
        )

    def test_plot(self):
        if show_plots:
            self.solution_species.plot(
                plot_components=True,
                plot_species=False,
                plot_total_concentration=False,
            )

            self.solution_species.plot(
                plot_components=False,
                plot_species=True,
                plot_total_concentration=False,
            )

            self.solution_species.plot(
                plot_components=False,
                plot_species=False,
                plot_total_concentration=True,
            )

            self.solution_species.plot(
                plot_components=True,
                plot_species=False,
                plot_total_concentration=True,
            )

            self.solution_species.plot(
                plot_components=True,
                plot_species=True,
                plot_total_concentration=True,
            )

            self.solution_species.plot(
                components=["A"],
                plot_components=True,
                plot_species=True,
                plot_total_concentration=True,
            )

            self.solution_species.plot(
                components=["B"],
                plot_components=True,
                plot_species=True,
                plot_total_concentration=True,
            )


if __name__ == "__main__":
    show_plots = True

    unittest.main()



from typing import Optional

import numpy as np

from CADETProcess import CADETProcessError
from CADETProcess.processModel import (
    BindingBaseClass,
    Cstr,
    FlowSheet,
    Inlet,
    LumpedRateModelWithoutPores,
    Outlet,
    Process,
)
from CADETProcess.simulator import Cadet


def simulate_solid_equilibria(
    binding_model: BindingBaseClass,
    buffer: list,
    unit_model: Optional[str] = "cstr",
    flush: Optional[list] = None,
    cadet_install_path: Optional[str] = None,
) -> list:
    """
    Simulate initial conditions for solid phase for given buffer.

    Parameters
    ----------
    binding_model : BindingBase
        Binding model describing relation between bulk and solif phase.
    buffer : list
        Buffer concentration in mM.
    unit_model : {'cstr', 'column'}, optional
        Unit model to be used in simulation. The default is 'cstr'.
    flush : list, optional
        Additional buffer for flushing column after loading.
        The default is None.
    cadet_install_path : str, optional
        Path to CADET installation. If None, the default CADET path is used.

    Raises
    ------
    CADETProcessError
        DESCRIPTION.

    Returns
    -------
    list
        Initial conditions for solid phase.
    """
    process_name = flow_sheet_name = "initial_conditions"
    component_system = binding_model.component_system

    # Unit Operations
    buffer_source = Inlet(component_system, name="buffer")
    buffer_source.c = buffer

    if flush is None:
        flush = buffer
    flush_source = Inlet(component_system, "flush")
    flush_source.c = flush

    if unit_model == "cstr":
        unit = Cstr(component_system, "cstr")
        unit.init_liquid_volume = 5e-7
        unit.const_solid_volume = 5e-7

        Q = 1e-6
        cycle_time = np.round(1000 * unit.volume / Q)
        unit.flow_rate = Q
    elif unit_model == "column":
        unit = LumpedRateModelWithoutPores(component_system, name="column")
        unit.length = 0.1
        unit.diameter = 0.01
        unit.axial_dispersion = 1e-6
        unit.total_porosity = 0.7

        Q = 60 / (60 * 1e6)
        cycle_time = np.round(10 * unit.volume / Q)
    else:
        raise CADETProcessError("Unknown unit model")

    try:
        q = binding_model.n_comp * binding_model.n_states * [0]
        capacity = binding_model.capacity
        if isinstance(capacity, list):
            capacity = capacity[0]

        q[0] = capacity
        unit.q = q
        c = binding_model.n_comp * [0]
        c[0] = buffer[0]
        unit.c = c
    except AttributeError:
        pass

    unit.binding_model = binding_model

    unit.solution_recorder.write_solution_bulk = True
    unit.solution_recorder.write_solution_solid = True

    outlet = Outlet(component_system, name="outlet")

    # flow sheet
    fs = FlowSheet(component_system, name=flow_sheet_name)

    fs.add_unit(buffer_source)
    fs.add_unit(flush_source)
    fs.add_unit(unit)
    fs.add_unit(outlet, product_outlet=True)

    fs.add_connection(buffer_source, unit)
    fs.add_connection(flush_source, unit)
    fs.add_connection(unit, outlet)

    # Process
    proc = Process(fs, name=process_name)
    proc.cycle_time = cycle_time

    # Create Events and Durations
    proc.add_event("buffer_on", "flow_sheet.buffer.flow_rate", Q)
    proc.add_event("buffer_off", "flow_sheet.buffer.flow_rate", 0, 0.9 * cycle_time)

    proc.add_event("eluent_off", "flow_sheet.flush.flow_rate", 0.0, 0.0)
    proc.add_event("eluent_on", "flow_sheet.flush.flow_rate", Q, 0.9 * cycle_time)

    # Simulator
    process_simulator = Cadet(cadet_install_path)
    proc_results = process_simulator.simulate(proc)

    if unit_model == "cstr":
        init_q = proc_results.solution[unit.name].solid.solution[-1, :]
    elif unit_model == "column":
        init_q = proc_results.solution[unit.name].solid.solution[-1, 0, :]

    init_q = np.round(init_q, 14)
    return init_q.tolist()



"""
=======================================
Plotting (:mod:`CADETProcess.plotting`)
=======================================

.. currentmodule:: CADETProcess.plotting

This module provides functionality for plotting in CADET-Process.

General Style
=============

.. autosummary::
    :toctree: generated/

    set_figure_style
    SecondaryAxis
    Layout
    set_layout

Setup Figure
============

.. autosummary::
    :toctree: generated/

    setup_figure
    create_and_save_figure


Annotations
===========

.. autosummary::
    :toctree: generated/

    Annotation
    add_annotations

Ticks
=====

.. autosummary::
    :toctree: generated/

    Tick
    set_yticks
    set_xticks

Fill Regions
============

.. autosummary::
    :toctree: generated/

    FillRegion
    add_fill_regions

Text
====

.. autosummary::
    :toctree: generated/

    add_text

Hlines
======

.. autosummary::
    :toctree: generated/

    HLines
    add_hlines

"""  # noqa

import sys
from functools import wraps
from typing import Any, Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib import cycler
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import (
    Callable,
    Integer,
    List,
    String,
    Structure,
    Tuple,
    UnsignedFloat,
)

this = sys.modules[__name__]

style = "medium"

color_dict = {
    "blue": matplotlib.colors.to_rgb("#000099"),
    "red": matplotlib.colors.to_rgb("#990000"),
    "green": matplotlib.colors.to_rgb("#009900"),
    "orange": matplotlib.colors.to_rgb("#D79B00"),
    "purple": matplotlib.colors.to_rgb("#896999"),
    "grey": matplotlib.colors.to_rgb("#444444"),
}
color_list = list(color_dict.values())
chromapy_cycler = cycler(color=color_list)

linestyle_cycler = cycler("linestyle", ["--", ":", "-."])

textbox_props = dict(facecolor="white", alpha=1)


figure_styles = {
    "small": {
        "width": 5,
        "height": 3,
        "linewidth": 2,
        "font_small": 8,
        "font_medium": 10,
        "font_large": 12,
        "color_cycler": chromapy_cycler,
    },
    "medium": {
        "width": 10,
        "height": 6,
        "linewidth": 4,
        "font_small": 20,
        "font_medium": 24,
        "font_large": 28,
        "color_cycler": chromapy_cycler,
    },
    "large": {
        "width": 15,
        "height": 9,
        "linewidth": 6,
        "font_small": 25,
        "font_medium": 30,
        "font_large": 40,
        "color_cycler": chromapy_cycler,
    },
}


def set_figure_style(style: Optional[str] = "medium") -> None:
    """
    Define the sytle of a plot.

    Can set the sytle of a plot for small, medium and large plots. The
    figuresize of the figure, the linewitdth and color of the lines and the
    size of the font can be changed by switching the style.

    Parameters
    ----------
    style : str, optional
        Style of a figure plot, default set to small.

    Raises
    ------
    CADETProcessError
        If no valid style has been chosen as parameter.
    """
    if style not in figure_styles:
        raise CADETProcessError("Not a valid style")

    width = figure_styles[style]["width"]
    height = figure_styles[style]["height"]
    linewidth = figure_styles[style]["linewidth"]
    font_small = figure_styles[style]["font_small"]
    font_medium = figure_styles[style]["font_medium"]
    font_large = figure_styles[style]["font_large"]
    color_cycler = figure_styles[style]["color_cycler"]

    plt.rcParams["figure.figsize"] = (width, height)
    plt.rcParams["lines.linewidth"] = linewidth
    plt.rcParams["font.size"] = font_small  # controls default text sizes
    plt.rcParams["axes.titlesize"] = font_small  # fontsize of the axes title
    plt.rcParams["axes.labelsize"] = font_medium  # fontsize of the x and y labels
    plt.rcParams["xtick.labelsize"] = font_small  # fontsize of the tick labels
    plt.rcParams["ytick.labelsize"] = font_small  # fontsize of the tick labels
    plt.rcParams["legend.fontsize"] = font_small  # legend fontsize
    plt.rcParams["figure.titlesize"] = font_large  # fontsize of the figure title
    plt.rcParams["axes.prop_cycle"] = color_cycler


set_figure_style()


def get_fig_size(
    n_rows: Optional[int] = 1,
    n_cols: Optional[int] = 1,
    style: Optional[str] = None,
) -> tuple[float, float]:
    """
    Get figure size for figures with multiple Axes.

    Parameters
    ----------
    n_rows : int, optional
        Number of rows in the figure. The default is 1.
    n_cols : int, optional
        Number of columns in the figure. The default is 1.
    style : str, optional
        Style to use for the figure. The default is None.

    Returns
    -------
    fig_size : tuple
        Size of the figure (width, height)
    """
    if style is None:
        style = this.style

    width = figure_styles[style]["width"]
    height = figure_styles[style]["height"]

    fig_size = (n_cols * width + 2, n_rows * height + 2)

    return fig_size


def setup_figure(
    n_rows: Optional[int] = 1,
    n_cols: Optional[int] = 1,
    style: Optional[str] = None,
    squeeze: Optional[bool] = True,
) -> tuple[Figure, Axes]:
    """
    Set up a matplotlib figure with specified dimensions and style.

    Parameters
    ----------
    n_rows : int, optional
        Number of rows in the figure, by default 1.
    n_cols : int, optional
        Number of columns in the figure, by default 1.
    style : str, optional
        Style to use for the figure. Uses a predefined style if None.
    squeeze : bool, optional
        If True, extra dimensions are removed from the returned Axes array, by default True.

    Returns
    -------
    tuple[Figure, Axes]
        A tuple containing the Figure object and an Axes object or an array of Axes objects.
    """
    if style is None:
        style = this.style
    set_figure_style(style)

    fig_size = get_fig_size(n_rows, n_cols)
    fig, axs = plt.subplots(
        nrows=n_rows, ncols=n_cols, squeeze=squeeze, figsize=fig_size
    )

    fig.tight_layout()

    return fig, axs


class SecondaryAxis(Structure):
    """Parameters for secondary axis."""

    components = List()
    y_label = String()
    y_lim = Tuple()
    transform = Callable()


class Layout(Structure):
    """General figure layout."""

    style = String()
    title = String()
    x_label = String()
    x_ticks = List()
    y_label = String()
    y_ticks = List()
    x_lim = Tuple()
    y_lim = Tuple()


def set_layout(
    ax: Axes,
    layout: Layout,
    show_legend: bool = True,
    ax_secondary: Optional[SecondaryAxis] = None,
    secondary_layout: Optional[Layout] = None,
) -> None:
    """
    Configure the layout of a matplotlib Axes object.

    Parameters
    ----------
    ax : Axes
        The primary matplotlib Axes object to configure.
    layout : Layout
        Layout object containing axis labels, limits, title, and ticks.
    show_legend : bool, optional
        Whether to display the legend. Default is True.
    ax_secondary : Optional[SecondaryAxis], optional
        The secondary Axes object, if applicable.
    secondary_layout : Optional[Layout], optional
        Layout object for the secondary axis, if applicable.
    """
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ax.set_xlabel(layout.x_label)
    ax.set_ylabel(layout.y_label)
    ax.set_xlim(layout.x_lim)
    ax.set_ylim(layout.y_lim)
    ax.set_title(layout.title)

    if layout.x_ticks is not None:
        set_xticks(layout.x_ticks)
    if layout.y_ticks is not None:
        set_yticks(layout.y_ticks)

    lines, labels = ax.get_legend_handles_labels()

    if ax_secondary is not None:
        ax_secondary.set_ylabel(secondary_layout.y_label)
        ax_secondary.set_ylim(secondary_layout.y_lim)

        if show_legend:
            lines_secondary, labels_secondary = ax_secondary.get_legend_handles_labels()
            ax_secondary.legend(
                lines_secondary + lines, labels_secondary + labels, loc=0
            )
    else:
        if show_legend and len(labels) != 0:
            ax.legend()


class Tick(Structure):
    """Parameters for Axes ticks."""

    location: Tuple()
    label: String()


def set_yticks(ax: Axes, y_ticks: list[Tick]) -> None:
    """
    Set the y-ticks on a matplotlib Axes object.

    Parameters
    ----------
    ax : Axes
        The matplotlib Axes object to set the y-ticks on.
    y_ticks : list[Tick]
        List of Tick objects containing location and label for each y-tick.
    """
    locs = np.array([y_tick.location for y_tick in y_ticks])
    labels = [y_tick.label for y_tick in y_ticks]
    ax.set_yticks(locs, labels)


def set_xticks(ax: Axes, x_ticks: list[Tick]) -> None:
    """
    Set the x-ticks on a matplotlib Axes object with rotation.

    Parameters
    ----------
    ax : Axes
        The matplotlib Axes object to set the x-ticks on.
    x_ticks : list[Tick]
        List of Tick objects containing location and label for each x-tick.
    """
    locs = np.array([x_tick.location for x_tick in x_ticks])
    labels = [x_tick.label for x_tick in x_ticks]
    plt.xticks(locs, labels, rotation=72, horizontalalignment="center")


def add_text(
    ax: Axes,
    text: str,
    position: tuple[float, float] = (0.05, 0.9),
    tb_props: Optional[Any] = None,
    **kwargs: Optional[dict],
) -> None:
    """
    Add text to a matplotlib Axes object.

    Parameters
    ----------
    ax : Axes
        The matplotlib Axes object to add text to.
    text : str
        The text to be added.
    position : tuple[float], optional
        The position of the text, default is (0.05, 0.9).
    tb_props : Optional[Any], optional
        Properties to update the textbox with.
    **kwargs : Optional[dict]
        Additional keyword arguments for text customization.
    """
    if tb_props is not None:
        textbox_props.update(tb_props)

    ax.text(
        *position,
        text,
        transform=ax.transAxes,
        verticalalignment="top",
        bbox=textbox_props,
        **kwargs,
    )


def add_overlay(
    ax: Axes,
    y_overlay: npt.ArrayLike | list[npt.ArrayLike],
    x_overlay: Optional[npt.ArrayLike] = None,
    **plot_args: Optional[dict],
) -> None:
    """
    Add overlay plot(s) to a matplotlib Axes object.

    Parameters
    ----------
    ax : Axes
        The matplotlib Axes object to which the overlay is added.
    y_overlay : npt.ArrayLike | list[npt.ArrayLike]
        The y-data for the overlay plot(s).
    x_overlay : Optional[list], optional
        The x-data for the overlay plot(s). If None, uses x-data from the first line in ax.
    **plot_args : Optional[dict]
        Additional keyword arguments for customizing the plot.
    """
    if not isinstance(y_overlay, list):
        y_overlay = [y_overlay]

    if x_overlay is None:
        x_overlay = ax.lines[0].get_xdata()

    for y_over in y_overlay:
        ax.plot(x_overlay, y_over, **plot_args)
        ax.set_prop_cycle(None)


class Annotation(Structure):
    """Parameters for text annotations."""

    text = String()
    xy = Tuple()
    xytext = Tuple()
    arrowstyle = "-|>"


def add_annotations(
    ax: Axes,
    annotations: list[Annotation],
) -> None:
    """Add list of annotations to axis ax."""
    for annotation in annotations:
        ax.annotate(
            annotation.text,
            xy=annotation.xy,
            xycoords="data",
            xytext=annotation.xytext,
            textcoords="offset points",
            arrowprops={
                "arrowstyle": annotation.arrowstyle,
            },
        )


class FillRegion(Structure):
    """Parameters for fill region."""

    color_index = Integer()
    start = UnsignedFloat()
    end = UnsignedFloat()

    y_max = UnsignedFloat()

    text = String()


def add_fill_regions(
    ax: Axes,
    fill_regions: list[FillRegion],
    x_lim: Optional[npt.ArrayLike] = None,
) -> None:
    """Add FillRegion to axes."""
    for fill in fill_regions:
        color = color_list[fill.color_index]
        ax.fill_between(
            [fill.start, fill.end],
            fill.y_max,
            alpha=0.3,
            color=color,
        )

        if fill.text is not None:
            if x_lim is None or fill.start < x_lim[0]:
                x_position = (x_lim[0] + fill.end) / 2
            else:
                x_position = (fill.start + fill.end) / 2
            y_position = 0.5 * fill.y_max

            ax.text(
                x_position,
                y_position,
                fill.text,
                horizontalalignment="center",
                verticalalignment="center",
            )


class HLines(Structure):
    """Parameters for plotting horizontal lines."""

    y = UnsignedFloat()
    x_min = UnsignedFloat()
    x_max = UnsignedFloat()


def add_hlines(ax: Axes, hlines: list[HLines]) -> None:
    """Add hlines to matplotlib Axes."""
    for line in hlines:
        ax.hlines(line.y, line.x_min, line.x_max)


def create_and_save_figure(func: Callable) -> Callable:
    """Wrap plot functions to provide some general utility."""

    @wraps(func)
    def wrapper(
        *args: Any,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        show: bool = True,
        file_name: Optional[str] = None,
        style: str = "medium",
        **kwargs: Any,
    ) -> tuple:
        """
        Wrap plot functions to provide some general utility.

        Parameters
        ----------
        *args :
            Parameters wrapped around.
        fig : Figure, optional
            Figure object.
        ax : Axes, optional
           Axes to plot on. If None, a new standard figure will be created.
        show : bool, optional
            If True, show plot. The default is False.
        file_name : str, optional
            Path for saving figure. If None, figure is not saved.
        style : str, optional
            Style for figure. Default i 'medium'.
        **kwargs :
            Additional parameters wrapped around.

        """
        if ax is None:
            fig, ax = setup_figure(style=style)

        func(*args, ax=ax, **kwargs)

        if fig is not None:
            fig.tight_layout()

        if file_name is not None:
            plt.savefig(file_name)

            plt.close(fig)
            if show:
                dummy = plt.figure(figsize=fig.get_size_inches())
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                plt.show()

        return fig, ax

    return wrapper
